"""
This module contains useful functions to help to generalize
common types and easy user deffined types into JSON

@author: VieVie31
"""
import json
import typing

from copy import deepcopy

__all__ = ["BonapityJSONEncoder"]

class NotRightType(Exception):
    pass

# Decorator
class ObjectToJSONizable:
    __all_transform_functions = []

    def __new__(cls, fun):
        """
        The meta-class construcor is the decorator indexing all possible
        transformations to be applied when `resolve` is called.

        The decored function should return a tuple containing the serialized
        object and a boolean (False by defaut) informing if the result is a
        case base (result JSON serializable).
        """
        ObjectToJSONizable.__all_transform_functions.append(fun)
        return fun

    @staticmethod
    def resolve(obj):
        """
        Try ton convert a given object into a JSON serializable
        object (typically `dict`) or die.

        <!> : This function use recursive call and so can be slow, 
        prefer making your object serializable yourself if possible.

        :param obj: the object to try to serialize

        :return: an object which is JSON serializable
        """
        # Copy the object to not destroy it
        obj = deepcopy(obj)

        # Find the right transformation
        out = NotRightType()
        stop_recurstion = False
        for transformation in ObjectToJSONizable.__all_transform_functions:
            try:
                out, stop_recurstion = transformation(obj)
                if not type(out) in [type(None), bool, int, float, str, list, dict]:
                    out = NotRightType()
                else:
                    break
            except:
                pass

        # No transform function succeedded to transform the object in dict
        if isinstance(out, NotRightType):
            raise NotRightType(f"Didn't succeed to JSON encode : {out}")
        else:
            # Try to encode it (for exemple numpy array may 
            # - somtimes - be directly encoded, no recursion needed)
            try:
                out, stop_recurstion = transform_default_json_encoder(out)
            except:
                pass
        # If allowed to stop recursion, return result as is
        if stop_recurstion:
            return out

        # Continue to make each
        if isinstance(out, dict):
            out = deepcopy(out)
            for k in out:
                out[k] = ObjectToJSONizable.resolve(out[k])
            return out
        elif isinstance(out, list):
            for i, v in enumerate(out):
                out[i] = ObjectToJSONizable.resolve(v)
            return out
        else:
            raise Exception(f"Type {type(out)} unexpected in resolve...")

# The transformation should be declared in the right order...
@ObjectToJSONizable
def transform_default_json_encoder(obj) -> typing.Tuple[typing.Dict, bool]:
    """
    Try to encode the object with the default python json encoder.
    If success, we do not need to do any more recursive call.
    This is a base case. 
    The First returned value can not be a dict but a JSON serializable object
    such as list or int.
    """
    return json.loads(json.dumps(obj)), True

@ObjectToJSONizable
def transform_python_dicts(_dict) -> typing.Tuple[typing.Dict, bool]:
    _dict = deepcopy(_dict)
    for k, v in _dict.items():
        _dict[k] = ObjectToJSONizable.resolve(v)
    return _dict, False 
    #TODO: do same function for defaultdict, mappingproxy, lists, etc.

@ObjectToJSONizable
def transform_get_object_state(obj) -> typing.Tuple[typing.Dict, bool]:
    return obj.__getstate__(), False

@ObjectToJSONizable
def transform_simple_object(obj) -> typing.Tuple[typing.Dict, bool]:
    return {**obj.__dict__}, False

@ObjectToJSONizable
def transform_numpy(ndarray) -> typing.Tuple[typing.List, bool]:
    return ndarray.tolist(), False

@ObjectToJSONizable
def transform_python_sets(_set) -> typing.Tuple[typing.List, bool]:
    return list(_set), False

#TODO: transform other generic types such as sets, dates
# or complex ones such as plots
    
class BonapityJSONEncoder(json.JSONEncoder):
    """
    This class inherit from json.JSONEncoder and can be used as bellow :
    ```python
    >>> json.dumps(myobject, cls=BonapityJSONEncoder)
    ```
    """
    def default(self, obj):
        return ObjectToJSONizable.resolve(obj)






"""
This module contains all the functions which will be converted into decorators.

@author: VieVie31, Ricocotam
"""
import json
import pickle
import urllib
import functools
import contextvars
import http.cookies

from .server import ThreadingBonAppServer, BonAppServer
from .decoration_classes import DecoratedFunctions, BonapityDecoratedFunction, BonapityException
from .json_encode import BonapityJSONEncoder
from .session import SessionManager

__all__ = ["bonapity", "vuosi"]


def vuosi(domain: str, port: int):
    """
    Decorator for python clients to simplify the requests data transfert
    and using the API like a bunch of blackbox functions...

    The decored function name should be a command of the api.
    Example:
    ```
    >>> @vuosi('localhost', 8888) # Specify the domain and the port
    >>> def myfun(a):
    ...     # Can create this function only if `/myfun?a=...` exists in the api
    ...     pass # No body needed, the decorator will write/execute is for you

    >>> myfun('coucou cici')
    ```
    """

    def inner_vuosi(fun):
        def fetch(**params):
            """
            All arguments are keyword arguments...
            For example do not write f(3) but f(x=3)...
            """
            params = pickle.dumps(params)
            r = urllib.request.urlopen(urllib.request.Request(
                f'http://{domain}:{port}/{fun.__name__}',
                data=params,
                headers={"Content-Type": "application/python-pickle"}
            ))
            if r.status >= 300:
                raise BonapityException(
                    f"Failed to fetch result, code : [{r.status}], message : {r.read()}"
                )
            res = r.read()
            # print(res)
            try:
                res = pickle.loads(res)
            except:
                res = json.loads(res.decode())
            return res

        fetch.__name__ == fun.__name__
        return fetch

    return inner_vuosi


class MetaBonAPIty(type):
    @property
    def cookies(cls):
        return cls._BonAPIty__cookies.get()
    
    @property
    def session(cls):
        return cls._BonAPIty__session.get()

class BonAPIty(object, metaclass=MetaBonAPIty):
    # Each function will have access to the current client cookies
    # with `BonAPIty.cookies.get()`. The content of the variable w'll be
    # changed updated with the contextvars
    __cookies = contextvars.ContextVar("bonapity.cookies")
    __session = contextvars.ContextVar("bonapity.session")

    @staticmethod
    def __exec_function(f, cookies=http.cookies.SimpleCookie(), session={}):
        global bonapity
        context = contextvars.copy_context()
        context.run(bonapity._BonAPIty__cookies.set, cookies)
        context.run(bonapity._BonAPIty__session.set, session)
        res = context.run(f)
        return res, context.get(bonapity.__cookies)

    @staticmethod
    def serve(
        port=8888, static_files_dir: str = './', index: str = None, 
        help: bool = True, timeout: int = 10, session_timeout: int = 600, 
        verbose: bool = True):
        """
        Serve your API forever.

        :param port:
            the port to serve the API
        :param static_files_dir:
            root of the directory to serve as static files
            (those files are served as GET only)
            prefer absolute path, less ambiguous (
                else depend of the current position of python
                ) => less problems
        :param index:
            if not None, use this page as index (returned as html)
        :param help:
            return the documentation of functions at
            `http://[SERVER]/help/[FUN_NAME|ROOT]`
        :param timeout:
            number of seconds before ending the function and returning
            timeout error message, if 0, no timeout is applied
        :param verbose:
            display some informations such as the port where the server w'll run

        :type port: int
        :type help: bool

        Example :
        ```
        >>> bonapity.serve(8080)
        Server running on  port : 8080
        ```
        """
        SessionManager.session_timeout = session_timeout

        PORT = port
        server_address = ("", PORT)
        handler = BonAppServer

        if verbose:
            print(f"Server running on  port : {PORT}")

        httpd = ThreadingBonAppServer(server_address, handler)

        httpd.RequestHandlerClass.bonapity = BonAPIty
        httpd.RequestHandlerClass.port = port

        httpd.RequestHandlerClass.static_files_dir = static_files_dir
        httpd.RequestHandlerClass.index = index
        httpd.RequestHandlerClass.help = help
        httpd.RequestHandlerClass.default_timeout = timeout

        httpd.RequestHandlerClass.cookies_context = contextvars.copy_context()
        

        httpd.serve_forever()

    @staticmethod
    def __new__(
        cls, fun=None, name: str = None, 
        timeout: int = None, mime_type: str = "auto", 
        json_encoder: json.JSONEncoder = BonapityJSONEncoder):
        """
        Get a simple HTTP GET API with this simple decorator.
        You'll be able to use your function at :
        `http://[SERVER]/[FUN_NAME|ALIAS_NAME]?[PARAMETERS]`.

        To start the server use the `serve` method.

        :param fun:
            the function we want to create a simple api
            it's safer to use type hints for input args
            we w'll cast the inputs for you
            the types should be python generics ones
        :param name:
            rename this function in the api to conflict names
        :param timeout:
            timeout to kill the function
        :param mime_type:
            specity your return mine-type if you want to return
            custom data such as binary images. If content-type
            given, the function is assumed returning byte data.
            If mine_type is set to "auto" (default) AND your
            function return type is `byte` we try to automatically
            detect the right mime type.
        :param json_encoder:
            the encoder to use when trying to serialize
            the return type as json, the default `BonapityJSONEncoder`
            is slower as it can handle more complex object than
            the default encoder, you can change it by the default
            `json.JSONEncoder` or your one encoder inheriting from it.
            Remember, if the encoder can't encode the object into
            a JSON format, the object returned will be pickled.

        Example:
        ```
        >>> from bonapity import bonapity

        >>> @bonapity      #or @bonapity('my_alias_fun_name')
        ... def add(a: int, b: int=0) -> int:
        ...     return a + b

        >>> if __name__ == '__main__':
        ...     bonapity.serve()
        ```
        """
        if fun is None:
            return functools.partial(
                BonAPIty.__new__, cls, name=name, timeout=timeout,
                mime_type=mime_type, json_encoder=json_encoder
            )
        elif type(fun) == str:
            return functools.partial(
                BonAPIty.__new__, cls, name=fun, timeout=timeout,
                mime_type=mime_type, json_encoder=json_encoder
            )

        fname = fun.__name__ if not name else name
        fname = f"{'' if fname[0] == '/' else '/'}{fname}"
        DecoratedFunctions.all[fname] = BonapityDecoratedFunction(
            fun, timeout, mime_type, json_encoder)

        @functools.wraps(fun)
        def f(*args, **kwargs):
            return fun(*args, **kwargs)

        return f


bonapity = BonAPIty



"""
This module contains the classes used internally
to store and manage the decorated functions.

@author: VieVie31
"""
import json
import typing

from dataclasses import dataclass

from .json_encode import BonapityJSONEncoder


class DecoratedFunctions(object):
    """
    This static class contains the listing of all functions decorated.
    """
    all = {}

    def __new__(cls):
        return DecoratedFunctions.all


@dataclass
class BonapityDecoratedFunction():
    """
    This class store the function and the timeout allowed for it.
    """
    fun: typing.Callable
    timeout: int = 0
    mime_type: str = "auto"
    json_encoder: json.JSONEncoder = BonapityJSONEncoder


class BonapityException(Exception):
    """
    Bonapity Exception, usefull for the `vuosi` decorator to differentiate
    exceptions from `bonapity` (server communication for example) from others.
    """
    pass



"""
This module contains classes help full to generate html page programmatically.

@author: VieVie31
"""
from typing import List

__all__ = ["make_html_index", "make_html_function_doc"]

class Tag:
    def __init__(self, *children, **attributes):
        self.children = children
        self.attributes = attributes

    @property
    def innerHTML(self):
        s = ""
        for c in self.children:
            s += str(s)

    def __str__(self):
        tag_name = self.__class__.__name__.lower()
        attributes = ' '.join([
            f"{a}='{self.attributes[a]}'"
            if isinstance(self.attributes[a], str) else
            f"{a}='{' '.join(map(str, self.attributes[a]))}'" 
            for a in self.attributes
        ])
        return f"<{tag_name} {attributes}>{''.join(map(str, self.children))}</{tag_name}>"

class ATag:
    def __init__(self, **attributes):
        self.attributes = attributes
    
    def __str__(self):
        tag_name = self.__class__.__name__.lower()
        attributes = ' '.join([
            f"{a}='{self.attributes[a]}'"
            if isinstance(self.attributes[a], str) else
            f"{a}='{' '.join(map(str, self.attributes[a]))}'" 
            for a in self.attributes
        ])
        return f"<{tag_name} {attributes}/>"

class FlattenStrTags:
    def __init__(self, *tags_objects):
        self.children = tags_objects
    
    def __str__(self):
        return ''.join(map(str, self.children))


class Html(Tag):
    pass

class Head(Tag):
    pass

class Title(Tag):
    pass

class Style(Tag):
    pass

class Body(Tag):
    pass

class Footer(Tag):
    pass

class H1(Tag):
    pass

class H2(Tag):
    pass

class H3(Tag):
    pass

class H4(Tag):
    pass

class H5(Tag):
    pass

class H6(Tag):
    pass

class P(Tag):
    pass

class Span(Tag):
    pass

class Div(Tag):
    pass

class Code(Tag):
    pass

class I(Tag):
    pass

class Tt(Tag):
    pass

class Script(Tag):
    pass

class A(Tag):
    pass

class Li(Tag):
    pass

class Ul(Tag):
    pass

class Details(Tag):
    pass

class Summary(Tag):
    pass

class Br(ATag):
    pass

class Img(ATag):
    pass

class Hr(ATag):
    pass


def make_html_index(modules) -> str:
    """
    Generate an HTML index page referencing all functions
    grouped by module in the API.

    :param modules: 
        defaultdict where keys are modules names
        and values list of function names
    """
    generated_index = str(Html(
        Head(Title("Index - List of all API's functions")),
        Body(
            H1("Index of functions available in the API"),
            Hr(),
            *[
                FlattenStrTags(
                    H2(m),
                    Ul(*[
                        Li(A(f, href=f"/help{f}"))
                        for f in modules[m]
                    ]),
                    Hr()
                )
                for m in modules
            ],
            Footer(
                'Generated by ★', A('bonAPIty', href='https://github.com/VieVie31/bonapity'), '★'
            )
        )
    ))
    return generated_index

def make_html_function_doc(fname, signature, docstring, python_code, js_code) -> str:
    """
    Generate the html documentation page for a given function
    """
    return str(Html(
        Head(Title(f"Documentation for : {fname}")),
        Body(
            H1(f"Documentation for : {fname}"),
            Div(H2("Signature"), I(Tt(signature))),
            Hr(),
            Div(
                H2("Description : "),
                Code(docstring, style='display:block;white-space:pre-wrap'),
                Hr()
            ),
            Div(
                H2("Generated Code"),
                "We help you by providing templates/wrappers to write the client using your API.",
                Details(
                    Summary("Python"),
                    I(Code(python_code, style='display:block;white-space:pre-wrap')),
                    "By using this decorator, all your arguments are becoming keywords arguments."
                ),
                Details(
                    Summary("Javascript"),
                    I(Code(js_code, style='display:block;white-space:pre-wrap'))
                ),
            ),
            Div(
                Hr(),
                Footer(
                    'Generated by ★', A('bonAPIty', href='https://github.com/VieVie31/bonapity'), '★'
                ),
                style="position:absolute;bottom:5;right:5;"
            )
        )
    ))



"""
This module contains useful function to infer mime-type.

@author: VieVie31
"""
from pathlib import Path
from collections import defaultdict


# Complete list : http://www.iana.org/assignments/media-types/media-types.xhtml
# TODO: parse this : https://www.freeformatter.com/mime-types-list.html
COMMON_MIME_TYPES = {
    '': 'application/octet-stream',
    '.': 'application/octet-stream',
    '.aac': 'audio/aac',
    '.abx': 'application/x-abiword',
    '.arc': 'application/octet-stream',
    '.avi': 'video/x-msvideo',
    '.azw': 'application/vnd.amazon.ebook',
    '.bin': 'application/octet-stream',
    '.bmp': 'image/bmp',
    '.bz': 'application/x-bzip',
    '.bz2': 'application/x-bzip2',
    '.csh': 'application/x-csh',
    '.css': 'text/css',
    '.csv': 'text/csv',
    '.doc': 'application/msword',
    '.doc': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    '.eot': 'application/vnd.ms-fontobject',
    '.epub': 'application/epub+zip',
    '.flac': 'audio/flac',
    '.gif': 'image/gif',
    '.htm': 'text/html',
    '.html': 'text/html',
    '.ico': 'image/x-icon',
    '.ics': 'text/calendar',
    '.jar': 'application/java-archive',
    '.jpeg': 'image/jpeg',
    '.jpg': 'image/jpeg',
    '.js': 'application/javascript',
    '.json': 'application/json',
    '.manifest': 'text/cache-manifest',
    '.mid': 'audio/midi',
    '.midi': 'audio/midi',
    '.mpeg': 'video/mpeg',
    '.mpgk': 'application/vnd.apple.installer+xml',
    '.mp3': 'audio/mpeg',
    '.mp4': 'video/mp4',
    '.odp': 'application/vnd.oasis.opendocument.presentation',
    '.ods': 'application/vnd.oasis.opendocument.spreadsheet',
    '.odt': 'application/vnd.oasis.opendocument.text',
    '.oga': 'audio/ogg',
    '.ogg': 'application/ogg',
    '.ogv': 'video/ogg',
    '.ogx': 'application/ogg',
    '.otf': 'font/otf',
    '.pdf': 'application/pdf',
    '.png': 'image/png',
    '.ppt': 'application/vnd.ms-powerpoint',
    '.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    '.rar': 'application/x-rar-compressed',
    '.rtf': 'application/rtf',
    '.sh': 'application/x-sh',
    '.svg': 'image/svg+xml',
    '.swf': 'application/x-shockwave-flash',
    '.tar': 'application/x-tar',
    '.tif': 'image/tiff',
    '.tiff': 'image/tiff',
    '.torrent': 'application/x-bittorrent',
    '.ts': 'application/typescript',
    '.ttf': 'font/ttf',
    '.vsd': 'application/vnd.visio',
    '.wav': 'audio/x-wav',
    '.weba': 'video/webm',
    '.webm': 'video/webm',
    '.webp': 'image/webp',
    '.woff': 'font/woff',
    '.woff2': 'font/woff2',
    '.xhtml': 'application/xhtml+xml',
    '.xls': 'application/vnd.ms-excel',
    '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    '.xml': 'application/xml',
    '.xul': 'application/vnd.mozilla.xul+xml',
    '.zip': 'application/zip',
    '.3gp': 'video/3gpp',  # Can also be audio/3gpp if no video file is present.
    '.3g2': 'video/3gpp2',  # Can also be audio/3gpp2 if no video file is present.
    '.7z': 'application/x-7z-compressed'
}


def extension_to_mime(extension: str) -> str:
    """
    Return mime-type associated to a given extension.
    If not found, return "application/octet-stream".

    :param extension: extension of the file formated as ".myextension"
    """
    global COMMON_MIME_TYPES

    if extension in COMMON_MIME_TYPES:
        return COMMON_MIME_TYPES[extension]
    else:
        return 'application/octet-stream'


def byte_to_mime(byte_data: bytes) -> str:
    """
    Return mime-type associated to a given bytes array.
    If not found, return "application/octet-stream".

    :param extension: extension of the file formated as ".myextension"
    """
    # https://en.wikipedia.org/wiki/List_of_file_signatures
    # RAR
    # 52 61 72 21 1A 07 00
    # 52 61 72 21 1A 07 01 00

    # Check in decreasing order of the magic number
    # byte length the potential application

    if byte_data[:11] == b'd8:announce':
        return 'application/x-bittorrent'

    f5_mime_type = {
        b'<html': 'text/html',
        b'<?xml': 'application/xml',
        b'%PDF-': 'application/pdf'
    }
    if byte_data[:5] in f5_mime_type:
        return f5_mime_type[byte_data[:5]]

    f4_mime_type = {
        b'\x89PNG': 'image/png',
        b'GIF8': 'image/gif',
        b'%PDF': 'application/pdf',
        b'\xff\xd8\xff\xdb': 'image/jpeg',
        b'\xff\xd8\xff\xe0': 'image/jpeg',
        b'\xff\xd8\xff\xee': 'image/jpeg',
        b'\xff\xd8\xff\xe1': 'image/jpeg',
        b'II*.': 'image/tiff',
        b'MM.*': 'image/tiff',
        b'PK\x03\x04': 'application/zip',
        b'PK\x05\x06': 'application/zip',
        b'\xca\xfe\xba\xbe': 'application/java-byte-code',
        b'OggS': 'application/ogg',
        b'fLaC': 'audio/flac',
        b'MThd': 'audio/rtp-midi',
        b'\x1aE\xdf\xa3': 'video/x-matroska',
        b'\x00asm': 'application/wasm',
        b'\x00\x00\x01\xb3': 'audio/mpeg',
        b'\x00\x00\x01\xba': 'audio/mpeg',
        b'AVI ': 'video/x-msvideo'
    }
    if byte_data[:4] in f4_mime_type:
        return f4_mime_type[byte_data[:4]]

    f3_mime_type = {
        b'ID3': 'audio/mpeg',
        b'RIF': 'video/x-msvideo'
    }
    if byte_data[:3] in f3_mime_type:
        return f3_mime_type[byte_data[:3]]

    # 42 4D #BMP
    # FF FB #MP3
    # 46 ?? ?? ?? ?? #AVI

    # Unknown magic number case
    return "application/octet-stream"


if __name__ == '__main__':
    pass



"""
This module contains all stuffs to handle the HTTP server.

@author: VieVie31
"""
import ast
import typing
import inspect
import functools
import html
import json
import base64
import pickle
import http.server
import http.cookies
import urllib.parse
import urllib.request
import socketserver
import os.path
import threading
import queue
import contextvars

from pathlib import Path
from copy import deepcopy
from collections import defaultdict

from .code_generation import generate_js, generate_python, generate_js_lib
from .decoration_classes import DecoratedFunctions
from .mime import extension_to_mime, byte_to_mime
from .html_generation import make_html_index, make_html_function_doc
from .session import SessionManager, make_session_id


def send_header(server_instance, code, content_type, cookies=None):#, session_id=None):
    """
    This function is used by the BonAppServer to send back the headers.

    :param server_instance:
        the instance of the BonAppServer (should call with `self`)
    :param code:
        the HTTP code status to return
    :param content_type:
        the HTTP `Content-Type` information (eg. txt/html, application/json)
    """
    server_instance.send_response(code)
    server_instance.send_header('Content-type', content_type)
    server_instance.send_header(
        "Access-Control-Allow-Origin",
        'null'
        if server_instance.headers["Origin"] is None
        else server_instance.headers["Origin"]
    )
    server_instance.send_header(
        'Access-Control-Allow-Methods',
        'GET, POST, PUT, DELETE, PATCH, OPTIONS'  # *?
    )

    # Send regular cookies
    if cookies != None:
        server_instance.send_header('Set-Cookie', cookies.output(header='', sep=';\r\nSet-Cookie: '))

    session_id = server_instance.get_cookies_and_session_id()[1]

    server_instance.send_header("Access-Control-Allow-Credentials", 'true')
    server_instance.send_header('Set-Cookie', f'BONAPITYSID={session_id if session_id else make_session_id()}')
    server_instance.end_headers()


class BonAppServer(http.server.BaseHTTPRequestHandler):
    def __init__(self, *args, **kargs):
        super(BonAppServer, self).__init__(*args, **kargs)
        self.help = True
        self.port = 80
        self.default_timeout = 1
        self.static_files_dir = None
        self.index = None
        self.bonapity = None

    def get_cookies_and_session_id(self):
        cookies = self.headers.get('Cookie')
        cookies = http.cookies.SimpleCookie(cookies)
        cookies = deepcopy(cookies)

        # BONAPITYSID is the reserved cookie name for sessions
        if "BONAPITYSID" in cookies:
            # Get the session id
            session_id = cookies["BONAPITYSID"].value
            # Do not allow modification of the cookie session, pop it
            cookies.pop("BONAPITYSID")
        else:
            session_id = make_session_id()
        return cookies, session_id

    def process(self, parameters, value_already_evaluated=False):
        """
        This method do the common stuffs between do_GET and do_POST.

        :param parameters: key-value dict of parameters (values are not parsed)
        """
        cookies, session_id = self.get_cookies_and_session_id()
        # Get the session
        session = SessionManager()[session_id]

        parsed_url = urllib.parse.urlparse(self.path)

        decorated_function = DecoratedFunctions.all[parsed_url.path]

        # Get the function in the decorated function list
        fun = decorated_function.fun
        sig = inspect.signature(fun)

        # Get the timeout informatin of the function, is None set a default one
        timeout = decorated_function.timeout
        if timeout is None:
            timeout = self.default_timeout

        # Get mime-type
        mime_type = decorated_function.mime_type

        # Check if parameters names matches
        # (and ignore default and *args, **kargs parameters)
        full_arg_spec = inspect.getfullargspec(fun)
        ignored_params_names = [
            k for k in sig.parameters.keys()
            if sig.parameters[k].default != inspect._empty
               or k == full_arg_spec.varargs
               or k == full_arg_spec.varkw
        ]

        if sorted(set(sig.parameters.keys()) - set(ignored_params_names)) \
                != sorted(set(parameters.keys()) - set(ignored_params_names)):
            send_header(self, 400, 'text/html')
            self.wfile.write(
                b"Parameters names do not match the signature of the function...")
            return

        # try
        for param_key, param_value in parameters.items():
            # print("param:", param_key, param_value)
            if len(param_value) != 1:
                send_header(self, 500, 'application/json')
                self.wfile.write(b"Each argument can be used only once !")
                return

            param_value = param_value[0]

            ftype = sig.parameters[param_key].annotation

            if ftype != str:
                # Try to evaluate
                evaluated = value_already_evaluated

                try:
                    # Evaluate as Python formated text
                    if not evaluated:
                        param_value = ast.literal_eval(param_value)
                        evaluated = True
                except:
                    pass

                try:
                    # Evalutate as JONS formated text
                    if not evaluated:
                        param_value = json.loads(param_value)
                        evaluated = True
                except:
                    pass

                # Return 500 error if failed to evaluate parameter
                if not evaluated:
                    send_header(self, 500, 'application/json')
                    self.wfile.write(
                        f"Parameter {param_key} : {param_value} {type(param_value)} poorly formated...".encode())
                    return

                # Try to cast
                if ftype == inspect._empty:
                    # No typing
                    # just affect the parsed value
                    parameters[param_key] = param_value
                else:
                    supported_generic_types = {
                        int: int(),
                        float: float(),
                        str: str(),
                        bool: bool(),
                        typing.List[(ftype.__args__[0],) if '__args__' in ftype.__dict__ else None]: list(),
                        typing.Tuple[(ftype.__args__[0],) if '__args__' in ftype.__dict__ else None]: tuple(),
                        typing.Set[(ftype.__args__[0],) if '__args__' in ftype.__dict__ else None]: set(),
                        typing.FrozenSet[(ftype.__args__[0],) if '__args__' in ftype.__dict__ else None]: frozenset(),
                        typing.Dict[ftype.__args__ if '__args__' in ftype.__dict__ and len(
                            ftype.__args__) == 2 else (None, None)]: dict()
                    }
                    if ftype in supported_generic_types:
                        # Cast the generic type
                        try:
                            parameters[param_key] = type(
                                supported_generic_types[ftype])(param_value)
                        except Exception as e:
                            send_header(self, 500, 'application/json')
                            self.wfile.write(
                                f"Error on parameter `{param_key}` : {e}".encode())
                            return
                    else:
                        # Not a generic python type :
                        # just affect the parsed value
                        parameters[param_key] = param_value
            else:
                # ftype was str
                parameters[param_key] = str(param_value)

        # Add the missing default parameters if not filled
        for k in sig.parameters.keys():
            if k not in parameters:
                parameters[k] = sig.parameters[k].default

        # Execute the function or die
        try:
            # Fill the first positional arguments in the right order
            f = functools.partial(
                fun,
                *[parameters[a] for a in inspect.getfullargspec(fun).args]
            )

            # Fill the *args if any
            if full_arg_spec.varargs is not None \
                and len(full_arg_spec.varargs):
                f = functools.partial(f, *parameters[full_arg_spec.varargs])

            # Fill the **kargs if any
            if full_arg_spec.varkw is not None \
                    and isinstance(parameters[full_arg_spec.varkw], dict) \
                    and len(full_arg_spec.varkw):
                f = functools.partial(f, **parameters[full_arg_spec.varkw])


            # Prepare the function to be executed within a context
            # Needed to resolve the contextvariables
            orig_f = f # Copy in another name to avoid infinite recursion
            f = lambda: self.bonapity._BonAPIty__exec_function(orig_f, cookies, session)

            if timeout > 0.:
                que = queue.Queue()
                thr = threading.Thread(target=lambda q: q.put(f()), args=(que,))
                thr.start()
                thr.join(timeout)

                # TODO: differentiate timeout form function crash

                if que.empty():
                    # Time out error
                    send_header(self, 500, 'text/html')
                    self.wfile.write(
                        f"""Timeout error: execution of {fun.__name__}
                        took more than the {timeout} allowed seconds
                        or function crashed... (see server logs)
                        """.encode()
                    )
                    return
                else:
                    # Get result from the top of the queue
                    res, cookies = que.get()
            else:
                # No timeout constraint
                res, cookies = f()

            # If a mime-type is given, return as is (byte data assumed)
            if mime_type not in [None, "auto"]:
                # Check if data is in byte format
                if type(res) != type(b''):
                    send_header(self, 500, 'text/html')
                    self.wfile.write(
                        f"{fun.__name__} didn't returned byte data but :{str(res)[:100]}...".encode())
                    return
                # The data are bytes
                send_header(self, 200, str(mime_type), cookies=cookies)
                self.wfile.write(res)
                return
            elif mime_type in [None, "auto"] and type(res) == type(b''):
                suffix = Path(parsed_url.path).suffix
                if suffix:
                    # If extension provided in file name detect from if
                    mime_type = extension_to_mime(suffix)
                else:
                    # Try to infer mine-type from 16 first bytes (enought)
                    # else return "application/octet-stream"
                    mime_type = byte_to_mime(res[:16])
                    # raise NotImplementedError()
                # The data are bytes
                send_header(self, 200, str(mime_type), cookies=cookies)
                self.wfile.write(res)
                return

            # Ecode result in JSON
            try:
                res = json.dumps(res, cls=decorated_function.json_encoder)

                # Send success
                send_header(self, 200, 'application/json', cookies=cookies)
                self.wfile.write(
                    f"{res}".encode()
                )
                return
            except Exception as e:
                # Encore into pickle and send a binary object
                try:
                    res = pickle.dumps(res)

                    # Send success
                    send_header(self, 200, 'application/python-pickle', cookies=cookies)
                    self.wfile.write(res)
                    return
                except Exception as nested_e:
                    e = nested_e
                send_header(self, 500, 'application/json')
                self.wfile.write(
                    f"TypeError on {fun.__name__} : {e}, result is not serializable (JSON nor pickle)... :'(".encode())
                return

        except Exception as e:
            send_header(self, 500, 'application/json')
            self.wfile.write(f"Error in {fun.__name__} : {str(e)}".encode())
            return

    def do_OPTIONS(self):
        # Send back informations for complex POST
        self.send_response(200, "ok")
        self.send_header('Access-Control-Allow-Methods',
                         self.headers["Access-Control-Request-Method"])
        self.send_header("Access-Control-Allow-Headers",
                         self.headers["Access-Control-Request-Headers"])
        self.send_header(
            "Access-Control-Allow-Origin",
            'null' if self.headers["Origin"] is None
            else self.headers["Origin"]
        )
        self.send_header("Access-Control-Allow-Credentials", 'true')
        self.end_headers()
        # <!> : this method should not contain anything more and no return !

    def __return_file(self, file_path, mime_type: str = None):
        """Serve a given file from a `path_file` and
        detect `mime_type` if not provided
        """
        # Serve the file
        try:
            with open(file_path, 'rb') as f:
                data = f.read()
                if mime_type is None: 
                    # Detect mime_type if `None`
                    mime_type = extension_to_mime(Path(file_path).suffix)
                    if mime_type == 'application/octet-stream':
                        # Try to determine mime from 16 first bytes
                        mime_type = byte_to_mime(data[:16])
                send_header(self, 200, mime_type)
                self.wfile.write(data)
                return
        except Exception as e:
            send_header(self, 500, 'text/html')
            self.wfile.write(
                f"Error in serving file {file_path}: {e}".encode())
            return

    def __return_help_page(self, parsed_url):
        "Return an html help index page or the help of a given function"
        fname = parsed_url.path[5:]

        # If no name given, make an index of all functions allowed in the API
        if fname in ['', '/', '/*']:
            modules = defaultdict(list)
            for fname in DecoratedFunctions.all.keys():
                modules[DecoratedFunctions.all[fname].fun.__module__].append(fname)

            html_out = make_html_index(modules)

            send_header(self, 200, 'text/html')
            self.wfile.write(html_out.encode())
            return
        elif not self.help and parsed_url.path == '/':
            send_header(self, 404, 'text/html')
            self.wfile.write(b'')
            return

        #FIXME: to allow regex
        elif fname not in DecoratedFunctions.all.keys():
            send_header(self, 404, 'text/html')
            self.wfile.write(
                f"{fname} : this function do not exists...".encode())
            return
        fun = DecoratedFunctions.all[fname].fun
        sig = html.escape(f"{fun.__name__}{inspect.signature(fun)}")
        doc = html.escape(fun.__doc__) if fun.__doc__ else ''

        # Extract domain, port from the client request (socket)
        domain, port = self.request.getsockname()

        html_out = make_html_function_doc(
            fname, sig, doc, 
            generate_python(fname, sig, doc, domain, port), 
            generate_js(fname, list(inspect.signature(fun).parameters.keys()), domain, port)
        )

        send_header(self, 200, 'text/html')
        self.wfile.write(html_out.encode())
        return
    
    def __extract_parameters_from_url(self, parsed_url):
        """Return the parameters extracted from the `parsed_url`
        as urlargs or base64 pickle and return a boolean indicating
        if the variables have been evaluated (True for pickle).
        This should be passed to the process method.

        :return: parameters as dict, boolean
        """
        # Will be replaced by True if data loaded from pickle
        value_already_evaluated = False

        urlargs = parsed_url.query.split('&')
        if len(urlargs) == 1 and urlargs[0].endswith('=' * urlargs[0].count('=')) and urlargs[0] != '':
            # Only an argument without value ?  `?Ym9uYXBpdHk=`
            # this should be a base64 pickled object
            try:
                parameters = base64.b64decode(urlargs[0])
                parameters = pickle.loads(parameters)
                if type(parameters) != dict or set(map(type, parameters)) != {str}:
                    send_header(self, 500, 'text/html')
                    self.wfile.write(
                        b"""
                        The encoded pickled object should be a dict (key: str, value)
                        with the name of argument as key and the corresponding
                        data as value, all keys should be strings...
                        """
                    )
                    return
                # Prepare parameters to be in the same format as
                # if processed by url : `{"key": [value(s)], ...}`
                parameters = {k: [parameters[k]] for k in parameters}
                value_already_evaluated = True
            except Exception as e:
                send_header(self, 500, 'text/html')
                self.wfile.write(str(e).encode())
                return
        else:
            # Regular way to pass data : `?arg1=val1&arg2=val2&...`
            parameters = urllib.parse.parse_qs(
                parsed_url.query
            )

        return parameters, value_already_evaluated

    def do_GET(self):
        parsed_url = urllib.parse.urlparse(self.path)

        # Serve the index page if exists else display an error
        if self.index and parsed_url.path in ['', '/']:
            self.__return_file(self.index, 'text/html')
            return
        # Check to display the js lib
        elif self.help and parsed_url.path.startswith('/help/') and parsed_url.query in ['lib=js', 'js']:
            send_header(self, 200, 'text/javascript')
            self.wfile.write(generate_js_lib(*self.request.getsockname()).encode())
            return
        # Check if display help
        elif self.help and (parsed_url.path.startswith('/help/') or parsed_url.path in ['', '/']):
            self.__return_help_page(parsed_url)
            return

        # If the function the user want to call do not exists
        # check if a file exists from the root `static_files_dir`
        #TODO: split the `?` and `#` ?
        file_path = (Path(self.static_files_dir) / Path(parsed_url.path[1:])).absolute()
        if parsed_url.path not in DecoratedFunctions.all.keys() and os.path.isfile(file_path):
            self.__return_file(file_path)
            return

        #FIXME: to allow regex
        elif parsed_url.path not in DecoratedFunctions.all.keys():
            send_header(self, 404, 'text/html')
            self.wfile.write(
                f"{parsed_url.path} : this function do not exists...".encode())
            return

        parameters, value_already_evaluated = self.__extract_parameters_from_url(parsed_url)

        self.process(parameters, value_already_evaluated)
        return

    def do_POST(self):
        content_len = int(self.headers['Content-Length'])
        rfile = self.rfile
        if self.headers['Content-Type'] == "application/json":
            try:
                parameters = json.loads(rfile.read(content_len).decode())
            except:
                send_header(self, 400, 'text/html')
                self.wfile.write(
                    b"""Content-Type : application/json expect json object...
                    we failed to load it, check your input...
                    """
                )
        elif self.headers['Content-Type'] == "application/python-pickle":
            try:
                parameters = pickle.loads(rfile.read(content_len))
            except:
                send_header(self, 400, 'text/html')
                self.wfile.write(
                    b"""Content type : application/python-pickle expect pickle binary object...
                    we failed to load it, check your input...
                    (maybe wrong version of pickle), we use python3...")
                    """
                )
        else:
            send_header(self, 400, 'text/html')
            self.wfile.write(
                f"""Bad Content-Type : {self.headers['Content-Type']},
                accepted CT are : application/json, application/python-pickle.
                """.encode())
            return
        # TODO: handle application/x-www-form-urlencoded ?
        # print(parameters)

        # Prepare paramters in the formats as if they was parsed from url
        parameters = {k: [parameters[k]] for k in parameters}

        self.process(parameters, value_already_evaluated=True)
        return

    def do_DELETE(self):
        self.do_GET()

    def do_PUT(self):
        self.do_POST()

    def do_PATCH(self):
        self.do_POST()


class ThreadingBonAppServer(socketserver.ThreadingMixIn, http.server.HTTPServer):
    """
    This class will make the server multi-threaded,
    allowing him to execute simulatneoulsy several client requests.
    """
    pass



"""
This module contains all the mechanic for handling user session.

@author: VieVie31
"""
import time
import random
import hashlib
import threading

from collections import defaultdict
from http.cookies import SimpleCookie


class UserSession:
    """
    User session have similar to dict as they support
    __setitem__ and __getitem___, but they also keep track of the timestamp
    of the last modifiction done via getter and setters.
    """
    def __init__(self):
        self.storage = {}
        self.creation_timestamp = time.time()
        self.last_timestamp = time.time()

    def __getitem__(self, key):
        self.last_timestamp = time.time()
        return self.storage[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.storage[self.__keytransform__(key)] = value
        self.last_timestamp = time.time()

    def __delitem__(self, key):
        del self.storage[self.__keytransform__(key)]
        self.last_timestamp = time.time()

    def __iter__(self):
        self.last_timestamp = time.time()
        return iter(self.storage)

    def __len__(self):
        self.last_timestamp = time.time()
        return len(self.storage)

    def __keytransform__(self, key):
        self.last_timestamp = time.time()
        return key
    
    def update_timestamp(self):
        self.last_timestamp = time.time()
    
    def __setattr__(self, name, value):
        if name in ['storage', 'sessionid', 'creation_timestamp', 'last_timestamp']:
            self.__dict__[name] = value
        else:
            raise Exception('attr setting not allowed')

class Singleton(type):
    _instances = {}
    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

class SessionManager(metaclass=Singleton):
    session_timeout = 600
    def __init__(self, session_timeout: int=600): #10min of inactivity before timeout
        """
        Create a session manager if not already existing.
        There is only one session manager instancied as this is a singleton.

        :param session_timeout:
            the session content of a specific user will be deleted
            after this number of second of inactivity (not using the session)
        """
        #raise NotImplementedError()
        SessionManager.session_timeout= session_timeout
        # `self.session` will contain for each SESSIONID as key a dict 
        # of the stored session data, aslo if the SESSIONID was not registered
        # the default dict corresponding to an emtpy session is {}
        self.sessions = defaultdict(UserSession) #(UserSession look like dict)
        # Create the thread doing the session garbage collector task
        threading.Timer(SessionManager.session_timeout, self.__clean_timedout_session).start()

    def __getitem__(self, sessionid) -> UserSession:
        return self.sessions[sessionid]
    
    def __setitem__(self, *a, **k):
        raise Exception('''
            __setattr__ is not allowed by SessionManager,
            you can only modify the returned UserSession')
        ''')
    
    def __delitem__(self, sessionid):
        del self.sessions[sessionid]
    
    def __clean_timedout_session(self):
        # Schedule the next GC call
        threading.Timer(
            1, # call each second ? too often ? I don't know...
            self.__clean_timedout_session
        ).start()
        # Delete if timeout or empty
        for sessionid in list(self.sessions.keys()):
            if time.time() - self.sessions[sessionid].last_timestamp > SessionManager.session_timeout \
                  or len(self.sessions[sessionid].storage) == 0:
                del self.sessions[sessionid]
    
def make_session_id() -> str:
    """
    This function will return a random session id as a sha1 hash 
    of the current timestamp concatened to a random number.
    """
    return hashlib.sha1(
        (f'{time.time()}' + f'{random.randint(0, 1000)}').encode()
    ).hexdigest()





from setuptools import setup

import bonapity

setup(
    name='bonapity',
    version=bonapity.__version__,
    description='Get a simple HTTP (REST) API with only this simple decorator : @bonapity !',
    long_description='bonAPIty is the simplest python (3.7) package allowing you to create API for your functions without writing any complicated line of server code ! See more [here](https://vievie31.github.io/bonapity/) !',
    long_description_content_type='text/markdown',
    url='https://vievie31.github.io/bonapity/',
    author='Olivier RISSER-MAROIX',
    author_email='orissermaroix@gmail.com',
    license='CC-BY',
    packages=['bonapity'],
    zip_safe=False
)





"""
This module contains all stuffs about the automatic code generation.

@author: VieVie31
"""
import inspect
import hashlib
import functools

from typing import List

from .decoration_classes import DecoratedFunctions

@functools.lru_cache(maxsize=1) # Build the lib once the cache it
def generate_js_lib(domain: str, port: int, js_lib_name: str = "bonapity_api") -> str:
    # Generate the name for the requestCORS method such as there'll 
    # be no conflit with any other name of the user lib.
    requestCORS_fname =  '_' + hashlib.sha1(b'requestCORS').hexdigest()[:8]
    all_fnames = list(map(lambda fname: fname[1:], DecoratedFunctions.all))
    salt = 0
    while requestCORS_fname in all_fnames:
        requestCORS_fname = '_' + hashlib.sha1(
            (''.join(all_fnames) + str(salt)).encode()
        ).hexdigest()[:8]
        salt += 1
    # This function will do all the AJAX and w'll be used by all the others
    requestCORS_function = """
        async """ + requestCORS_fname + """(method, url, data={}) {
            var xhr = new XMLHttpRequest();
            return new Promise(function(resolve, reject) {
            xhr.onreadystatechange = function() {
                if (xhr.readyState == 4) {
                    if (xhr.status >= 300)
                    reject('[' + xhr.status + '] : ' + xhr.statusText);
                    else
                    resolve(JSON.parse(xhr.responseText));
                }
                }
                xhr.open(method, url, true);
                xhr.withCredentials = true;
                xhr.setRequestHeader('Content-Type', 'application/json');
                xhr.send(JSON.stringify(data));
            });
        }
    """
    # Adding attributes for the domain name 
    # and the port and generating all functions
    out_lib = f""" let {js_lib_name} = {{ 
        {requestCORS_function}, 
        domain: "{domain}",
        port: {port},
        {(',' + chr(10)).join([
            f'''
            async {fname[1:]}({', '.join(inspect.signature(fun.fun).parameters.keys())}) {{
                keys = [{', '.join(map(lambda x: f'"{x}"', inspect.signature(fun.fun).parameters.keys()))}];
                if (arguments.length > keys.length)
                    throw new Exception('Too many arguments...');
                params = {{}};
                for (var i = 0; i < arguments.length; i++)
                    params[keys[i]] = arguments[i];
                url = 'http://' + this.domain + ':' + this.port + '/{fname[1:]}';
                return await this.{requestCORS_fname}('POST', url, params);
            }}
            '''
            for fname, fun in DecoratedFunctions.all.items()
        ])}
        
        }};"""
    return out_lib


def generate_js(fname: str, pnames: List[str], domain: str, port: int) -> str:
    fname = fname[1:]  # remove the starting '/'
    return f"""
    // Include 'http://{domain}:{port}/help/?lib=js' in your html header
    // All wrapper functions are `async` and should be called with `await`
    """


def generate_python(fname: str, signature: str, doc: str, domain: str, port: int) -> str:
    # doc = '' #f"""'''{doc.replace("'''", '"' * 3)}\'\'\'"""
    # Replace the default function name in the signature...
    signature = '('.join([fname[1:]] + signature.split('(')[1:])
    return f"""
    from bonapity import vuosi

    @vuosi('{str(domain)}', {port})
    def {signature}:\n\tpass
    """.replace("    ", '')



"""
Get a simple HTTP (REST) API with only this simple decorator : `@bonapity` !
See documentation with : `help(bonapity)`.

@author: VieVie31
"""
from .decorators import bonapity, vuosi

__version_info__ = (0, 1, 19)
__version__ = '.'.join(map(str, __version_info__))
__author__ = "Olivier RISSER-MAROIX (VieVie31)\nAdrien POUYET (Ricocotam)"

__all__ = ["bonapity", "vuosi"]
